import numpy as np

# This simple example solves the following constrainted optimization problem:
#
# argmin_{x} \sum_{i = 0}^{4} x_{i}
# s. t. 61 / x_{0}^3
#     + 37 / x_{1}^3
#     + 19 / x_{2}^3
#     + 7 / x_{3}^3
#     + 1 / x_{4}^3
#     - 1 < 0

# The API of the optimizer is entirely defined in the conveks module
import conveks

# Let us define the optimization problem, i.e. the objective,
# constraints as well as their derivative with respect to the
# design variables.

def get_objective(xFromOpti):
    objective = 0.0624*np.sum(xFromOpti)
    constraints = np.array([61.0/(xFromOpti[0]**3.0)\
        + 37.0/(xFromOpti[1]**3.0)\
        + 19.0/(xFromOpti[2]**3.0)\
        + 7.0/(xFromOpti[3]**3.0)\
        + 1.0/(xFromOpti[4]**3.0)\
        - 1.0])
    return objective,constraints

def get_grad(xFromOpti):
    numVariables = len(xFromOpti)
    grad_objective = 0.0624*np.ones(numVariables)
    grad_constraints = np.zeros(numVariables * 1)
    grad_constraints[0] = -3.0*61.0/(xFromOpti[0]**4.0)
    grad_constraints[1] = -3.0*37.0/(xFromOpti[1]**4.0)
    grad_constraints[2] = -3.0*19.0/(xFromOpti[2]**4.0)
    grad_constraints[3] = -3.0*7.0/(xFromOpti[3]**4.0)
    grad_constraints[4] = -3.0*1.0/(xFromOpti[4]**4.0)
    return grad_objective,grad_constraints

# Maximum number of iterations
maxIter = 10000

# number of variables and constraints
numVariables = 5

# initial value of the variables in the design space
initialPoint = 5.0*np.ones(numVariables)
# initialPoint = np.minimum(initialPoint,9.0)
# initialPoint = np.maximum(initialPoint,2.0)

# lower bound for design variables (all set to 1)
lowerBound = np.ones(numVariables)

# upper bound for design variables (all set to 10)
upperBound = 10.0 * np.ones(numVariables)

# Initialize conveks
conveks.initialize()

# Initialize the MMA optimizer
mma = conveks.mma.initialize(initialPoint, lowerBound, upperBound)

# Set some options for MMA
conveks.mma.option.setNumber('General.Verbosity', 0)
conveks.mma.option.setNumber('General.SaveHistory', 0)
conveks.mma.option.setNumber('SubProblem.asymptotesRmax', 100)
conveks.mma.option.setNumber('SubProblem.asymptotesRmin', 0.001)
conveks.mma.option.setNumber('SubProblem.isContinuation', 0)
# conveks.mma.option.setNumber('SubProblem.cInitial', 100)
# conveks.mma.option.setNumber('SubProblem.cFinal', 10000)
# conveks.mma.option.setNumber('SubProblem.cIteration', 50)
# conveks.mma.option.setNumber('SubProblem.isContinuation', 1)
# conveks.mma.option.setNumber('SubProblem.cMultiply', 10)
conveks.mma.option.setNumber('General.SaveHistory', 0)
# conveks.mma.option.setNumber("Solver.IterationsNonLinearMax", 200)
# conveks.mma.option.setNumber("Solver.IterationsNonLinearRelaxMax", 50)
conveks.mma.option.show()

# Get iteration count (here it will be 1 - could be different in case of restart)
it = conveks.mma.getOuterIteration()
kkt_norm = 1.0

# get (copy of) current point
x = conveks.mma.getCurrentPoint()

# compute objective function and constraints
# as well as their sensitivity with respect to design variables at `x'
objective,constraints = get_objective(x)
grad_objective,grad_constraints = get_grad(x)

print('iter.   obj.   max(constr.)   L2-norm(kkt)   NL. loops   point')
print('%3i  %2.4e  %2.4e  %2.4e  %3i'%(it-1, objective, np.max(np.abs(constraints)), kkt_norm, 0),x)

while it <= maxIter:

    # call MMA and update the current point
    conveks.mma.updateCurrentPoint(objective,grad_objective,constraints,grad_constraints)

    # get the count of outer iterations
    it = conveks.mma.getOuterIteration()

    # get (copy of) current point
    x = conveks.mma.getCurrentPoint()

    # compute objective function and constraints
    # as well as their sensitivity with respect to design variables at `x'
    objective,constraints = get_objective(x)
    grad_objective,grad_constraints = get_grad(x)

    # evaluate the L2-norm of KKT
    kkt_norm = conveks.mma.getKKTNorm(constraints,grad_objective,grad_constraints)

    # and get the total number of nonlinear loops
    nlloops = conveks.mma.countNonlinearIterations()

    print('%3i  %2.4e  %2.4e  %2.4e  %3i'%(it-1, objective, np.max(np.abs(constraints)), kkt_norm, nlloops),x)

# show the problem at the converged state
print('*'*80)
print('Optimization has converged after %i iterations.'%(it-1))
print('  > point           :',x)
print('  > objective       : %.4e'%(objective))
print('  > max(constraints): %.4e'%(np.max(np.abs(constraints))))
print('  > L2-norm of kkt  : %.4e'%(kkt_norm))
print('*'*80)

# This should be called at the end of MMA
conveks.mma.finalize()

# This should be called at the end
conveks.finalize()
